mod vm;

use std::{env, fs};
use std::io::{stderr, Write};
use std::sync::Arc;
use vm::FootVM;

use pixels::{Pixels, SurfaceTexture};

use winit::application::ApplicationHandler;
use winit::dpi::LogicalSize;
use winit::event::{StartCause, WindowEvent};
use winit::event_loop::{ActiveEventLoop, ControlFlow, EventLoop};
use winit::window::{Window, WindowAttributes, WindowButtons, WindowId};

const WIDTH: u32 = 64;
const HEIGHT: u32 = 64;
const BITS_PER_PIXEL: u32 = 4;
const PIXELS_PER_ADDRESS: u32 = 16 / BITS_PER_PIXEL;
const MEMORY_SLOTS: u32 = WIDTH * HEIGHT / PIXELS_PER_ADDRESS;
const BASE_ADDRESS: u32 = 0xFFFF - MEMORY_SLOTS;
const PIXEL_SCALE: u32 = 8;

pub struct App
{
    vm: FootVM,
    window: Option<Arc<Window>>,
    pixels: Option<Pixels<'static>>
}

impl App
{
    fn new(vm: FootVM) -> Self
    {
        App
        {
            vm,
            window: None,
            pixels: None
        }
    }

    fn update(&mut self)
    {
        self.vm.cycle();
    }

    fn draw(&mut self)
    {
        let pixels = &mut self.pixels.as_mut().unwrap();
        let frame = pixels.frame_mut();

        for ptr in 0..MEMORY_SLOTS
        {
            let mut value = self.vm[(ptr + BASE_ADDRESS) as u16];

            for k in 0..PIXELS_PER_ADDRESS
            {
                let pixel_value = ((value & 0xF) * 17) as u8;
                let pixel_index = ((ptr * PIXELS_PER_ADDRESS + k) * 4) as usize;
                frame[pixel_index + 0] = pixel_value;
                frame[pixel_index + 1] = pixel_value;
                frame[pixel_index + 2] = pixel_value;
                frame[pixel_index + 3] = 0xFF;

                value >>= BITS_PER_PIXEL;
            }
        }

        pixels.render().expect("");
    }
}

impl ApplicationHandler for App
{
    fn new_events(&mut self, _: &ActiveEventLoop, cause: StartCause)
    {
        match cause
        {
            StartCause::Poll =>
                {
                    self.update();
                    self.draw()
                },
            _ => {}
        }
    }

    fn resumed(&mut self, event_loop: &ActiveEventLoop)
    {
        let window = event_loop.create_window(
            WindowAttributes::default()
                .with_resizable(false)
                .with_enabled_buttons(WindowButtons::CLOSE | WindowButtons::MINIMIZE)
                .with_inner_size(LogicalSize::new(WIDTH * PIXEL_SCALE, HEIGHT * PIXEL_SCALE))
                .with_title("foot vm")
        ).unwrap();
        let window = Arc::new(window);

        self.window = Some(window);

        let surface_texture = SurfaceTexture::new(WIDTH * PIXEL_SCALE, HEIGHT * PIXEL_SCALE, self.window.as_ref().unwrap().clone());
        self.pixels = Some(Pixels::new(WIDTH, HEIGHT, surface_texture).unwrap());
    }

    fn window_event(&mut self, event_loop: &ActiveEventLoop, _: WindowId, event: WindowEvent)
    {
        match event
        {
            WindowEvent::RedrawRequested => self.draw(),
            WindowEvent::CloseRequested => event_loop.exit(),
            _ => {}
        }
    }
}

fn main()
{
    let args: Vec<String> = env::args().collect();
    if args.len() < 2
    {
        _ = stderr().write("No filename supplied".as_bytes())
    }
    else
    {
        let src = fs::read_to_string(&args[1])
            .expect("Unable to read file");

        let mut vm = FootVM::new();
        vm.load(src.as_str(), Some(&mut std::io::stderr())).expect("");

        let event_loop = EventLoop::new().unwrap();
        event_loop.set_control_flow(ControlFlow::Poll);

        let mut app = App::new(vm);
        event_loop.run_app(&mut app).expect("");
    }
}
